from fmtr.tools import version_tools as version

__version__ = version.read()


import fmtr.tools.dataclass_tools as dataclass
import fmtr.tools.datatype_tools as datatype
import fmtr.tools.environment_tools as env
import fmtr.tools.environment_tools as environment
import fmtr.tools.function_tools as function
import fmtr.tools.hash_tools as hash
import fmtr.tools.import_tools as import_
import fmtr.tools.iterator_tools as iterator
import fmtr.tools.json_tools as json
import fmtr.tools.path_tools as path
import fmtr.tools.platform_tools as platform
import fmtr.tools.random_tools as random
import fmtr.tools.string_tools as string
import fmtr.tools.name_tools as name
import fmtr.tools.logging_tools as logging
import fmtr.tools.async_tools as asyncio
import fmtr.tools.packaging_tools as packaging
from fmtr.tools.logging_tools import logger

from fmtr.tools.import_tools import MissingExtraMockModule

from fmtr.tools.constants import Constants

# Submodules
from fmtr.tools.path_tools import Path, PackagePaths, AppPaths
from fmtr.tools import ai_tools as ai

import fmtr.tools.setup_tools as setup
from fmtr.tools.setup_tools import Setup, SetupPaths, Dependencies, Tools

try:
    from fmtr.tools import augmentation_tools as augmentation
except ImportError as exception:
    augmentation = MissingExtraMockModule('augmentation', exception)

try:
    from fmtr.tools import yaml_tools as yaml
except ImportError as exception:
    yaml = MissingExtraMockModule('yaml', exception)

try:
    from fmtr.tools import docker_tools as docker
    from fmtr.tools.docker_tools import Container
except ImportError as exception:
    docker = Container = MissingExtraMockModule('docker.api', exception)

try:
    from fmtr.tools import parallel_tools as parallel
except ImportError as exception:
    parallel = MissingExtraMockModule('parallel', exception)

try:
    from fmtr.tools import profiling_tools as profiling
    from fmtr.tools.profiling_tools import Timer
except ImportError as exception:
    profiling = Timer = MissingExtraMockModule('profiling', exception)

try:
    import fmtr.tools.process_tools as process
    from fmtr.tools.process_tools import ContextProcess
except ImportError as exception:
    process = ContextProcess = MissingExtraMockModule('process', exception)

try:
    from fmtr.tools import tokenization_tools as tokenization
except ImportError as exception:
    tokenization = MissingExtraMockModule('tokenization', exception)

try:
    from fmtr.tools import unicode_tools as unicode
except ImportError as exception:
    unicode = MissingExtraMockModule('unicode', exception)

try:
    from fmtr.tools import netrc_tools as netrc
except ImportError as exception:
    netrc = MissingExtraMockModule('netrc', exception)

try:
    from fmtr.tools import spaces_tools as spaces
except ImportError as exception:
    spaces = MissingExtraMockModule('spaces', exception)

try:
    from fmtr.tools import hfh_tools as hfh
except ImportError as exception:
    hfh = MissingExtraMockModule('hfh', exception)

try:
    from fmtr.tools import merging_tools as merging
    from fmtr.tools.merging_tools import merge
except ImportError as exception:
    merging = merge = MissingExtraMockModule('merging', exception)

try:
    from fmtr.tools import api_tools as api
except ImportError as exception:
    api = MissingExtraMockModule('api', exception)

try:
    from fmtr.tools import data_modelling_tools as dm
except ImportError as exception:
    dm = MissingExtraMockModule('dm', exception)

try:
    from fmtr.tools import json_fix_tools as json_fix
except ImportError as exception:
    json_fix = MissingExtraMockModule('json_fix', exception)

try:
    from fmtr.tools import semantic_tools as semantic
except ImportError as exception:
    semantic = MissingExtraMockModule('semantic', exception)

try:
    from fmtr.tools import metric_tools as metric
except ImportError as exception:
    metric = MissingExtraMockModule('metric', exception)

try:
    from fmtr.tools import html_tools as html
except ImportError as exception:
    html = MissingExtraMockModule('html', exception)

try:
    from fmtr.tools import interface_tools as interface
except ImportError as exception:
    interface = MissingExtraMockModule('interface', exception)

try:
    from fmtr.tools import openai_tools as openai
except ImportError as exception:
    openai = MissingExtraMockModule('openai', exception)

try:
    from fmtr.tools import google_api_tools as google_api
except ImportError as exception:
    google_api = MissingExtraMockModule('google.api', exception)

try:
    from fmtr.tools import caching_tools as caching
except ImportError as exception:
    caching = MissingExtraMockModule('caching', exception)

try:
    from fmtr.tools import pdf_tools as pdf
except ImportError as exception:
    pdf = MissingExtraMockModule('pdf', exception)

try:
    from fmtr.tools import tabular_tools as tabular
except ImportError as exception:
    tabular = MissingExtraMockModule('tabular', exception)

try:
    from fmtr.tools import debugging_tools as debug
except ImportError as exception:
    debug = MissingExtraMockModule('debug', exception)

try:
    from fmtr.tools import settings_tools as sets
except ImportError as exception:
    sets = MissingExtraMockModule('sets', exception)

try:
    from fmtr.tools import pattern_tools as patterns
except ImportError as exception:
    patterns = MissingExtraMockModule('patterns', exception)

try:
    from fmtr.tools import dns_tools as dns
except ImportError as exception:
    dns = MissingExtraMockModule('dns', exception)

try:
    from fmtr.tools import http_tools as http
    from fmtr.tools.http_tools import Client
except ImportError as exception:
    http = Client = MissingExtraMockModule('http', exception)



import dns
import httpx
import socket
from dns.message import Message, QueryMessage


class Server:

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def resolve(self, message: Message) -> QueryMessage:
        raise NotImplemented

    def start(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind((self.host, self.port))
        print(f"Listening on {self.host}:{self.port}")
        while True:
            data, addr = sock.recvfrom(512)
            question = dns.message.from_wire(data)
            answer = self.resolve(question)
            sock.sendto(answer.to_wire(), addr)


class Proxy(Server):
    url = "https://cloudflare-dns.com/dns-query"
    headers = {"accept": "application/dns-json"}

    def resolve(self, question: Message) -> Message:
        qname = question.question[0].name.to_text()
        qtype = dns.rdatatype.to_text(question.question[0].rdtype)
        params = {"name": qname, "type": qtype}

        response = httpx.get(self.url, headers=self.headers, params=params)
        response.raise_for_status()

        answer = dns.message.make_response(question)
        answer.flags |= dns.flags.RA

        data = response.json()
        for answer_rr in data.get("Answer", []):
            name = dns.name.from_text(answer_rr["name"])
            rtype = answer_rr["type"]
            ttl = answer_rr["TTL"]
            rdata = dns.rdata.from_text(dns.rdataclass.IN, rtype, answer_rr["data"])
            rrset = dns.rrset.from_rdata(name, ttl, rdata)
            answer.answer.append(rrset)

        return answer


if __name__ == "__main__":
    proxy = Proxy("127.0.0.1", 5354)
    proxy.start()



import sys
from datetime import datetime
from functools import cached_property
from itertools import chain
from typing import List, Dict

from fmtr.tools.constants import Constants
from fmtr.tools.path_tools import Path
from fmtr.tools.path_tools.path_tools import FromCallerMixin


class SetupPaths(FromCallerMixin):
    """

    Canonical paths for a repo.

    """

    def __init__(self, path=None, org=Constants.ORG_NAME):

        """

        Use calling module path as default path, if not otherwise specified.

        """
        if not path:
            path = self.from_caller()

        self.org_name = org
        self.repo = Path(path)

    @property
    def readme(self):
        return self.repo / 'README.md'

    @property
    def version(self):
        return self.path / Constants.FILENAME_VERSION

    @cached_property
    def path(self):

        from fmtr.tools import setup

        if self.org:
            base = self.org
        else:
            base = self.repo

        directories = [base / dir for dir in setup.find_packages(base)]

        if len(directories) != 1:
            dirs_str = ', '.join([str(dir) for dir in directories])
            raise ValueError(f'Expected exactly one directory in {self.repo}, found {dirs_str}')

        package = next(iter(directories))
        return package

    @property
    def org(self):
        if not self.org_name:
            return False
        org = self.repo / self.org_name
        if not org.is_dir():
            return False
        return org

    @property
    def is_namespace(self) -> bool:
        return bool(self.org)

    @property
    def name(self) -> str:
        return self.path.stem


class Setup(FromCallerMixin):
    AUTHOR = 'Frontmatter'
    AUTHOR_EMAIL = 'innovative.fowler@mask.pro.fmtr.dev'

    REQUIREMENTS_ARG = 'requirements'
    SKIP_DIRS = {'data', 'build', 'dist', '.*', '*egg-info*'}

    def __init__(self, dependencies, paths=None, org=Constants.ORG_NAME, console_scripts=None, client=None, do_setup=True, **kwargs):

        self.kwargs = kwargs

        if type(dependencies) is not Dependencies:
            dependencies = Dependencies(**dependencies)
        self.dependencies = dependencies

        requirements_extras = self.get_requirements_extras()

        if requirements_extras:
            self.print_requirements()
            return

        self.org = org

        if not paths:
            paths = SetupPaths(path=self.from_caller(), org=self.org)
        self.paths = paths

        self.client = client
        self.console_scripts = console_scripts

        if do_setup:
            self.setup()

    def get_requirements_extras(self):
        if self.REQUIREMENTS_ARG not in sys.argv:
            return None

        extras_str = sys.argv[-1]
        extras = extras_str.split(',')
        return extras

    def print_requirements(self):
        reqs = []
        reqs += self.dependencies.install

        for extra in sys.argv[-1].split(','):
            reqs += self.dependencies.extras[extra]
        reqs = '\n'.join(reqs)
        print(reqs)



    def get_entrypoint_path(self, key, value):
        if value:
            return f'{self.name}.{value}:{key}'
        else:
            return f'{self.name}:{key}'

    @property
    def entrypoints(self):
        if self.console_scripts:
            return dict(
                console_scripts=[f'{key} = {self.get_entrypoint_path(key, value)}' for key, value in self.console_scripts.items()],
            )
        else:
            return dict()

    @property
    def name(self):
        if self.paths.is_namespace:
            return f'{self.paths.org_name}.{self.paths.name}'
        return self.paths.name

    @property
    def author(self):
        if self.client:
            return f'{self.AUTHOR} on behalf of {self.client}'
        return self.AUTHOR

    @property
    def copyright(self):
        if self.client:
            return self.client
        return self.AUTHOR

    @property
    def long_description(self):

        return self.paths.readme.read_text()

    @property
    def version(self):
        return self.paths.version.read_text().strip()

    @property
    def find(self):

        from fmtr.tools import setup

        if self.paths.is_namespace:
            return setup.find_namespace_packages
        else:
            return setup.find_packages

    @property
    def packages(self):

        excludes = list(self.SKIP_DIRS) + [f'{name}.*' for name in self.SKIP_DIRS if '*' not in name]

        packages = self.find(where=str(self.paths.repo), exclude=excludes)
        return packages

    @property
    def package_dir(self):
        if self.paths.is_namespace:
            return {'': str(self.paths.repo)}
        else:
            return None

    @property
    def package_data(self):
        return {self.name: [Constants.FILENAME_VERSION]}

    @property
    def url(self):
        return f'https://github.com/{self.org}/{self.paths.name}'

    @property
    def data(self):
        data = dict(
            name=self.name,
            version=self.version,
            author=self.author,
            author_email=self.AUTHOR_EMAIL,
            url=self.url,
            license=f'Copyright © {datetime.now().year} {self.copyright}. All rights reserved.',
            long_description=self.long_description,
            long_description_content_type='text/markdown',
            packages=self.packages,
            package_dir=self.package_dir,
            package_data=self.package_data,
            entry_points=self.entrypoints,
            install_requires=self.dependencies.install,
            extras_require=self.dependencies.extras,
        ) | self.kwargs
        return data

    def setup(self):

        from fmtr.tools import setup

        return setup.setup_setuptools(**self.data)


class Tools:
    MASK = f'{Constants.LIBRARY_NAME}[{{extras}}]'

    def __init__(self, *extras):
        self.extras = extras

    def __str__(self):
        extras_str = ','.join(self.extras)
        return self.MASK.format(extras=extras_str)



class Dependencies:
    ALL = 'all'
    INSTALL = 'install'

    def __init__(self, **kwargs):
        self.dependencies = kwargs

    def resolve_values(self, key) -> List[str]:
        """

        Flatten a list of values.

        """
        values_resolved = []
        values = self.dependencies[key]

        for value in values:
            if value == key or value not in self.dependencies:
                # Add the value directly if it references itself or is not a dependency key.
                values_resolved.append(str(value))
            else:
                # Recurse into nested dependencies.
                values_resolved += self.resolve_values(value)

        return values_resolved

    @cached_property
    def extras(self) -> Dict[str, List[str]]:
        """

        Flatten dependencies.

        """
        resolved = {key: self.resolve_values(key) for key in self.dependencies.keys()}
        resolved.pop(self.INSTALL, None)
        resolved[self.ALL] = list(set(chain.from_iterable(resolved.values())))
        return resolved

    @cached_property
    def install(self):
        return self.resolve_values(self.INSTALL)


if __name__ == '__main__':
    ds = Dependencies(
        install=['version', 'yaml'],

        yaml=['yamlscript', 'pyyaml'],
        logging=['logfire', 'version'],
        version=['semver', 'av'],
        av=['av']
        # Add the rest of your dependencies...
    )

    ds

    setup = Setup(
        # client='Acme',
        dependencies=ds,
        description='some tools test',
        console_scripts=dict(
            cache_hfh='console_script_tools',
            test=None,
        )
    )
    data = setup.get_data_setup()
    data



from dataclasses import dataclass
from functools import cached_property
from typing import List

import regex as re

from fmtr.tools.logging_tools import logger


class RewriteCircularLoopError(Exception):
    """

    Circular loop error

    """


@dataclass
class Rewrite:
    """
    Represents a single rule for pattern matching and target string replacement.

    This class is used to define a rule with a pattern and a target string.
    The `pattern` is a regular expression used to identify matches in input text.
    The `target` allows rewriting the identified matches with a formatted string.
    It provides properties for generating a unique identifier for use as a regex group name and compiling the provided pattern into a regular expression object.

    """
    pattern: str
    target: str

    @cached_property
    def id(self):
        """
        
        Regex group name.
        
        """
        return f'id{abs(hash(self.pattern))}'

    @cached_property
    def rx(self):
        """

        Regex object.

        """
        return re.compile(self.pattern)

    def apply(self, match: re.Match):
        """

        Rewrite using the target string and match groups.

        """
        target = self.target.format(**match.groupdict())
        return target


@dataclass
class Rewriter:
    """
    
    Represents a Rewriter class that handles pattern matching, rule application, and text rewriting.
    Compiles a single regex pattern from a list of rules, and determines which rule matched.
    It supports initialization from structured rule data, execution of a single rewrite pass, and
    recursive rewriting until a stable state is reached.

    """
    rules: List[Rewrite]

    @cached_property
    def pattern(self):
        """
        
        Provides a dynamically generated regex pattern based on the rules provided.

        """
        patterns = [fr"(?P<{rule.id}>{rule.pattern})" for rule in self.rules]
        sorted(patterns, key=len, reverse=True)
        pattern = '|'.join(patterns)
        return pattern

    @cached_property
    def rule_lookup(self):
        """

        Dictionary mapping rule identifiers to their corresponding rules.
        """

        return {rule.id: rule for rule in self.rules}

    @cached_property
    def rx(self):
        """

        Regex object.

        """
        return re.compile(self.pattern)

    def rewrite_pass(self, source: str):
        """

        Single rewrite pass.
        Rewrites the provided source string based on the matching rule.

        """

        match = self.rx.fullmatch(source)

        if not match:
            return source

        match_ids = {k: v for k, v in match.groupdict().items() if v}
        match_id = match_ids & self.rule_lookup.keys()

        if len(match_id) != 1:
            msg = f'Multiple group matches: {match_id}'
            raise ValueError(msg)

        match_id = next(iter(match_id))
        rule = self.rule_lookup[match_id]
        target = rule.apply(match)

        logger.debug(f'Rewrote using {match_id=}: {source=} -> {target=}')

        return target

    def rewrite(self, source: str) -> str:
        """

        Rewrites the provided text by continuously applying rewrite rules until no changes are made
        or a circular loop is detected.

        """
        history = []
        previous = source

        def get_history_str():
            return ' -> '.join(history)

        with logger.span(f'Rewriting "{source}"...'):
            while True:
                if previous in history:
                    history.append(previous)
                    msg = f'Loop detected on node "{previous}": {get_history_str()}'
                    raise RewriteCircularLoopError(msg)

                history.append(previous)

                new = previous

                new = self.rewrite_pass(new)

                if new == previous:
                    break

                previous = new

            logger.debug(f'Finished rewriting: {get_history_str()}')

        return previous

    @classmethod
    def from_data(cls, data):
        rules = [Rewrite(*pair) for pair in data.items()]
        self = cls(rules=rules)
        return self



